"""Backward-compatible route exports."""

from . import (  # noqa: F401
    analytics,
    app_server,
    base,
    file_chat,
    flows,
    messages,
    repos,
    review,
    sessions,
    settings,
    shared,
    system,
    usage,
    voice,
)

__all__ = [
    "analytics",
    "app_server",
    "base",
    "file_chat",
    "flows",
    "messages",
    "repos",
    "review",
    "sessions",
    "settings",
    "shared",
    "system",
    "usage",
    "voice",
]
